/**
 * Clase MyWorld.
 * @author Ivn Hopenhayn
 */
import javax.swing.Timer;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.Graphics2D;
import java.util.*;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.BasicStroke;
import java.awt.geom.Line2D;

public class MyWorld implements ActionListener {
    /**
 * Tiempo de refresco de la pantalla.
 */
	public static final double DEAFAUL_REFRESH = 0.003;  // 30 [ms]
/**
 * Tiempo en que se actualizan los estados.
 */
	public static final double DEFAUL_DELTA = 0.0001;    // 1 [ms]
/**
 *  Ancho de pantalla
 */
	public static int WIDTH = 600;  // in pixels
/**
 *  Largo de pantalla
 */
	public static int HEIGHT = 500; // in pixels
/**
 *  Origen del eje x.
 */
	public static int X_ORIGEN = (int)(WIDTH*0.1); // MyWorld space origin (x,y) will be on
/**
 * Origen del eje y.
 */
	public static int Y_ORIGEN = (int)(HEIGHT*0.9); // (X_ORIGEN,Y_ORIGEN) of the panel space.
/**
 * tranforma ejes coordenados.
 */
	public static AffineTransform SPACE_TRANSFORM;  // transforms (x,y) in (X,Y) of panel.
/**
 * tranformacion inversa de ejes coordenados.
 */
	public static AffineTransform SPACE_INVERSE_TRANSFORM; // transforms (X,Y) in (x,y) of my world.
/**
 * linea en 2 dimensiones.
 */
	public static Line2D.Double SCALE_LINE;  // 1 [m] line
/**
 * escala de ejes, 1 metro son 100 pixeles.
 */
	private static double AXES_SCALE = 100.0;  // 1 [m] equals 100 [pixels]

	static {
		SPACE_TRANSFORM = AffineTransform.getTranslateInstance(X_ORIGEN, Y_ORIGEN);
		SPACE_TRANSFORM.scale(AXES_SCALE,-AXES_SCALE);  // it also inverts direction of y-coordinate 
		try {
			SPACE_INVERSE_TRANSFORM = SPACE_TRANSFORM.createInverse();
		}catch (NoninvertibleTransformException e){ 
			System.exit(0);
		}
		SCALE_LINE = new Line2D.Double(300,15, 300+AXES_SCALE,15);  // 1[m] line.
	}
/**
 * tiempo de simulacin.
 */
	private double t;              // Simulation time
/**
 * tiempo de refresco en simulacin
 */
	private Timer pase;            // Clock to refresh the simulation's view. 
/**
 * delta t para actualizacin de estados
 */
	private double delta_t;        // in seconds
/**
 * perodo de refresco
 */
	private double refreshPeriod;  // in seconds
/**
 * gravedad
 */
	private Vector2D g;
/**
 * viscosidad del aire
 */
	private float viscosity;
/**
 * panel
 */
	private PhysicsLabPanel labPanel;
/**
 * panel grfico
 */
   private EnergyPlotPanel plotPanel;

/**
 * vector de elementos fisicos
 */
	private Vector<PhysicsElement> elements;
	private SerialTest serial;

    /** 
 * Constructor.  inicializa MyWorld 
 */
	public MyWorld(){
		t = 0;
		refreshPeriod = 0.04;      // initial value 40 [ms]
		delta_t = 0.001;          // initial value 1 [ms]
		pase = new Timer((int)(refreshPeriod*1000), this);
		g = new Vector2D(0,0);  // start with no gravity
		viscosity = 0.0f;        // no friction
		elements = new Vector<PhysicsElement>();
		labPanel = null;
		plotPanel = null;
	}
    /**
 * Setea el panel.  
 * @param p panel de clase PhysicsLabPanel que setea labPanel
 */
	public void setPanel (PhysicsLabPanel p) {
		labPanel = p;
	}
	public void setPanel (EnergyPlotPanel p) {              
      plotPanel = p;
   }
   public double getT(){
      return t;
   }
/**
 * Setea Delta t
 * @param delta tiempo de muestreo
 */
	public void setDelta_t(double delta) {
		delta_t = delta;
	}
		public SerialTest getSerial()
	{
		return serial;
	}
	public void setSerial (SerialTest s) {
		serial = s;
	}
/**
 * Setea perodo de refresco 
 * @param rp parmetro que setea la variable refreshPeriod
 */
	public void setRefreshPeriod (double rp) {
		refreshPeriod = rp;
		pase.setDelay((int)(refreshPeriod*1000)); // convert from [s] to [ms]
	}
/**
 * Setea gravedad
 * @param gravity  elemento de clase Vector2d que setea g
 */
	public void setGravity (Vector2D gravity) {
		g = gravity; 
	}
	        public double getKEnergy(){													// Se obtiene energia cintica del conjunto
            double KEnergy=0;
            for (PhysicsElement e: elements)  
				if (e.getEliminado() == false){                                  // como los objetos existen, se debe preguntar si estan eliminados.
                if(e instanceof Ball)                                         // para cada elemento se identifica si es bola y
                    KEnergy+= 0.5*e.getMass()*((e.getSpeed()).module())*((e.getSpeed()).module());} // suma la Energía Cinética
            return KEnergy;
	}
        /************************************/
        public double getPEnergy(){														// energia potencial del conjunto
            double PEnergy=0;
            for (PhysicsElement e: elements)  
				if (e.getEliminado() == false){                              // para cada elemento
						if(e instanceof Ball)
					 	PEnergy+=g.module()*e.getMass()*(e.getPosition()).getY();}  // suma la Energa Cintica y Potencial
            return PEnergy;
        }
        /***********************************/
        public double getMEnergy(){                 // Suma la energía potencial y la cinética
            return getKEnergy()+getPEnergy();       // para todos los elementos
	}
        /***********************************/
   

/**
 * Obtiene vector de gravedad
 * @return g  vector2D de gravedad
 */
	public Vector2D getGravity(){
		return g;
	}
/**
 * Setea viscosidad del aire
 * @param v  setea variable viscosity
 */
	public void setViscosity(float v){
		viscosity = v;
	}
/**
 * Obtiene viscosidad del aire
 * @return viscosidad del aire
 */
	public float getViscosity(){
		return viscosity;
	}
/**
 * Suma elementos al vector que agrupa PhysicsElement
 * @param e  Elemento fsico a sumar
 */

	public void addElement(PhysicsElement e) {
		elements.add(e);
		if (labPanel!=null)
			labPanel.repaint();
	}
/**
 * Comienza la simulacin
 */
	public void start() {
		if(pase.isRunning()) return;
		labPanel.desableMouseListener();
		pase.start();      
	}
/**
 * Detiene simulacin
 */
	public void stop(){
		pase.stop();
		labPanel.enableMouseListener();
	}
/**
 * Computa prximo estado y los actualiza.
 * @param event hace funcionar el mtodo cuando ocurre algo de tipo ActionEvent.
 */
	public void actionPerformed (ActionEvent event) {  // simulate time passing
		double nextStop=t+refreshPeriod;
		while (t<nextStop) {
			for (PhysicsElement e: elements)   // for each element
				e.computeNextState(delta_t, this);  // compute next state
			for (PhysicsElement e: elements)    // for all elements
				e.updateState();                // update its state
			t+=delta_t;
		}
		labPanel.repaint();	
		plotPanel.repaint(); 	
	}
/**
 * Retorna elemento con el cual se choca
 * @param anchor elemento fsico con el que se podra chocar
 * @return e  Slo si choca retorna el objeto con el que se choca
 */
	public PhysicsElement getNearCollision (PhysicsElement anchor) { // first implementation
		for (PhysicsElement e: elements) 
			if ((anchor != e) && anchor.collide(e))
				return e;
		return null;      
	}
/**
 * Calcula si hay choque
 * @param anchor elemento fsico de choque
 * @return nearest  elemento con el cual se est chocando
 */
	public PhysicsElement getNearestCollision (PhysicsElement anchor) {
		double minDistance=Double.MAX_VALUE;
		double distance;
		PhysicsElement nearest=null;
		for (PhysicsElement e: elements) {
			if (anchor!=e) {
				distance = anchor.getVectorBetweenCenters(e).module()
				-anchor.getRadius()-e.getRadius();
				if ((distance< minDistance) && anchor.collide(e)){
					minDistance = distance;
					nearest = e;
				}
			}
		}
		return nearest;      
	}
/**
 * Pinta componentes
 * @param g2d elemento que pinta
 */
    public void paintComponent(Graphics2D g2d){
	//g2d.drawString("ELO329 1er.Sem. 2012,   1 [m] = ", 85, 15);
	//g2d.draw(SCALE_LINE);
	g2d.transform(SPACE_TRANSFORM);
	g2d.setStroke(new BasicStroke(0.02f));
	for (PhysicsElement e: elements)
	    if (e instanceof Drawable)
		((Drawable)e).draw(g2d);
    }
/**
 * Muestra si un punto est dentro de un elemento grfico  
 * @param d coordenada x del punto
 * @param f coordenada y del punto
 * @return de  retorna elemento grfico que contiene al punto
 */
    public Drawable find(double d, double f) {
	for (PhysicsElement e: elements)
	    if (e instanceof Drawable){
		Drawable de = (Drawable) e;
		if (de.contains(d,f)) return de;
	    }
	return null;
	
    }
    
} 
